/*! \file mmc.c \brief MultiMedia and SD Flash Card Interface. */
//*****************************************************************************
//
// File Name	: 'mmc.c'
// Title		: MultiMedia and SD Flash Card Interface
// Author		: Pascal Stang - Copyright (C) 2004
// Created		: 2004.09.22
// Revised		: 2004.09.22
// Version		: 0.1
// Target MCU	: Atmel AVR Series
// Editor Tabs	: 4
//
// NOTE: This code is currently below version 1.0, and therefore is considered
// to be lacking in some functionality or documentation, or may not be fully
// tested.  Nonetheless, you can expect most functions to work.
//
// This code is distributed under the GNU Public License
//		which can be found at http://www.gnu.org/licenses/gpl.txt
//
//*****************************************************************************

//----- Include Files ---------------------------------------------------------
#include <avr/io.h>			// include I/O definitions (port names, pin names, etc)
#include <avr/interrupt.h>	// include interrupt support
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "global.h"		// include our global settings
#include "Xspi.h"		// include spi bus support
#include "task.h"
#include "MessageQueue.h"
#include "curses.h"
#include "io.h"
#include "cio.h"
#include "utils.h"

#include "Xmmc.h"
#include "ff.h"

// include project-specific hardware configuration
#include "mmcconf.h"

// Global variables
void MMCTask(void *arg)  __attribute__ ( ( noreturn ) );

MESSAGE_QUEUE *SDcardMSG;
TCB Tmmc;
static WINDOW *mmcW;

// Functions

void mmcInit(void)
{
	// initialize SPI interface
	// release chip select
	XSPI_CLR_CE();
	CreateTask(&Tmmc,MMCTask,512,20,"T1tsk",NULL);
	Insert(&ActiveTasks,&Tmmc);
	SDcardMSG = MqInit(32,"SDCardMsg");
}

int mmcResponse(unsigned char response)
{
	unsigned short count = 0xfff;
	unsigned char result;
	int loop = 1;
	while((count > 0) && loop)
	{
		result = XspiTransferByte(0xff);
		if(result == response)
			loop = 0;
		--count;
	}
	if(count == 0)
		return 1;
	else
		return 0;
}


u08 mmcReset(void)
{
	u16 retry;
	u08 r1=0;
	int i;
	char loop;
	char *buff;

	buff = malloc(128);


	// send dummy bytes with CS high before accessing
	wprintw(mmcW,"Dummy\n");
	for(i=0;i<10;++i) 
		buff[i] = 0xff;
	XspiTransfer(buff,i);

	XSPI_SET_CE();	//CS On
	wprintw(mmcW,"CE ON\n");
	XspiTransferByte(0xff);
	i = 0;
	buff[i++] = MMC_GO_IDLE_STATE | 0x40;
	buff[i++] = 0;
	buff[i++] = 0;
	buff[i++] = 0;
	buff[i++] = 0;
	buff[i++] = 0x95;	// crc valid only for MMC_GO_IDLE_STATE
	XspiTransfer(buff,i);

	wprintw(mmcW,"Check Resp\n");
	if(mmcResponse(0x01) == 1)
	{
		// release chip select
		XSPI_CLR_CE();
		wprintw(mmcW,"IDLE ERROR\n");
		return 1;
	}
	loop = 1;
	retry = 256;
	do
	{
		i = 0;
		buff[i++] = 0xff;
		buff[i++] = MMC_APP_CMD | 0x40;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0x95;	// crc valid only for MMC_GO_IDLE_STATE
		XspiTransfer(buff,i);

		if(mmcResponse(0x01) == 1)
		{
			// release chip select
			XSPI_CLR_CE();
			wprintw(mmcW,"APP_CMD\n");
			return 1;
		}

		i = 0;
		buff[i++] = 0xff;
		buff[i++] = MMC_SEND_OP_COND_A | 0x40;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0;
		buff[i++] = 0x95;	// crc valid only for MMC_GO_IDLE_STATE
		XspiTransfer(buff,i);


		if(mmcResponse(0x00) == 0)
		{
			// release chip select
			loop = 0;
		}
		else
			TimeDelay(5);
		--retry;
	}while(loop && retry);
	if(retry == 0) return 1;

	i = 0;
	buff[i++] = 0xff;
	buff[i++] = MMC_SET_BLOCKLEN | 0x40;
	buff[i++] = 0;
	buff[i++] = 0;
	buff[i++] = 2;
	buff[i++] = 0;
	buff[i++] = 0x95;	// crc valid only for MMC_GO_IDLE_STATE
	XspiTransfer(buff,i);

	if(mmcResponse(0x00) == 1)
	{
		// release chip select
		XSPI_CLR_CE();
		wprintw(mmcW,"BL ERROR\n");
		return 1;
	}
	XSPI_CLR_CE();
	free(buff);
	return r1;
}



unsigned short mmcSendStatus(void)
{
	union {
		unsigned short rV;
		unsigned char cV[2];
	} R;
	char *b = malloc(64);

	// assert chip select
	XSPI_SET_CE();
	R.cV[0] = mmcCommand(MMC_SEND_STATUS, 0,b);
	R.cV[1] = XspiTransferByte(0xff);
	// release chip select
	XSPI_CLR_CE();
	free(b);
	return R.rV;
}

u08 mmcSendCommand(u08 cmd, u32 arg)
{
	u08 r1;
	char *b = malloc(64);
	// assert chip select
	XSPI_SET_CE();
	// issue the command
	r1 = mmcCommand(cmd, arg,b);
	// release chip select
	XSPI_CLR_CE();
	free(b);
	return r1;
}

u08 mmcRead(u32 sector, u08* buffer)
{
	u08 r1;
	u16 i;

	wprintw(mmcW,"R Sector %ld->%04x\n",sector,(int)buffer);
	char *b = malloc(128);

	for(i=0;i<10;++i)
		b[i] = 0xff;
	XspiTransfer(b,i);
	
	// assert chip select
	XSPI_SET_CE();
	// issue command
	r1 = mmcCommand(MMC_READ_SINGLE_BLOCK, sector<<9,b);
	#ifdef MMC_DEBUG
	rprintf("MMC Read Block R1=0x%x\r\n", r1);
	#endif
	// check for valid response
	if(r1) wprintw(mmcW,"R1 = %02x\n",r1 & 0x0ff);
	if(r1 != 0x00)
		return r1;
	// wait for block start
	while(XspiTransferByte(0xFF) != MMC_STARTBLOCK_READ);
	// read in data
	for(i=0; i< 512;++i) buffer[i] = 0xff;
	XspiTransfer((char *)buffer,512);
/*
	for(i=0; i<0x200; i++)
	{
		*buffer++ = XspiTransferByte(0xFF);
	}
*/
	// read 16-bit CRC
	XspiTransferByte(0xFF);
	XspiTransferByte(0xFF);
	// release chip select
	XSPI_CLR_CE();
	free(b);
	// return success
	return 0;
}

u08 mmcWrite(u32 sector, u08* buffer)
{
	u08 r1;
	char *b = malloc(64);

	wprintw(mmcW,"W Sector %ld <- %04x\n",sector,buffer);
	// assert chip select
	XSPI_SET_CE();
	// issue command
	r1 = mmcCommand(MMC_WRITE_BLOCK, sector<<9,b);
	#ifdef MMC_DEBUG
	rprintf("MMC Write Block R1=0x%x\r\n", r1);
	#endif
	// check for valid response
	if(r1 != 0x00)
		return r1;
	// send dummy
	XspiTransferByte(0xFF);
	// send data start token
	XspiTransferByte(MMC_STARTBLOCK_WRITE);
	// write data

	XspiTransfer((char *)buffer,512);
	// write 16-bit CRC (dummy values)
	XspiTransferByte(0xFF);
	XspiTransferByte(0xFF);
	// read data response token
	r1 = XspiTransferByte(0xFF);
	if( (r1&MMC_DR_MASK) != MMC_DR_ACCEPT)
		return r1;
	#ifdef MMC_DEBUG
	rprintf("Data Response Token=0x%x\r\n", r1);
	#endif
	// wait until card not busy
	while(!XspiTransferByte(0xFF));
	// release chip select
	XSPI_CLR_CE();
	// return success
	free(b);
	return 0;
}

u08 mmcCommand(u08 cmd, u32 arg, char *b)
{
	u08 r1;
	u08 retry=0;
	// send command
	int i;
	i = 0;

	b[i++] = 0xff;
	b[i++] = cmd | 0x40;
	b[i++] = arg>>24;
	b[i++] = arg>>16;
	b[i++] = arg>>8;
	b[i++] = arg;
	b[i++] = 0x95;	// crc valid only for MMC_GO_IDLE_STATE
	XspiTransfer(b,i);

	// end command
	// wait for response
	// if more than 8 retries, card has timed-out
	// return the received 0xFF
	while((r1 = XspiTransferByte(0xFF)) == 0xFF)
		if(retry++ > 8) break;
	// return response
	return r1;
}

void SecDump(int fd, char *b)
{
	char *s = malloc(256);
	int i,j,l;
	unsigned adr=0;

	for(j=0;j<512;j+=16)
	{
		l = 0;
		l += UnsignedToHex(adr,&s[l],' ');
		for(i=0;i<16;++i)
		{
			l += CharToHex(b[i+j],&s[l],(i==7)?' ':'-');
		}
		for(i=0;i<16;++i)
		{
			if(isprint(b[i+j]))
				s[l++] = b[i+j];
			else
				s[l++] = '.';
		}
		s[l++] = '\r';
		s[l++] = '\n';
		_write(fd,s,l);
		adr += 16;
	}
	free(s);
}

void MMCTask(void* arg)
{
	WINDOW *win;
	unsigned char v;
	MSG *msg;
	char *b;
	int i;
	unsigned short uS;
	unsigned int uI;
	unsigned long *pL;
	unsigned short *pU;
	int fd;
	unsigned char *sbuff;

	extern void wHexDump(WINDOW *w,char *b,int n);
	extern void HexDump(int ,char *,int);
	FIL *FileObj;
	FATFS *FSYS;

	//setupt display window
	win = newwin(16,30,16,30,NULL,FG(YELLOW) | BG(BLUE),WIN_BOXED);
	scrollok(win,1);
	mmcW = win;
	WinSetTitle(win, "SD CARD");		//set window title
	b = malloc(64);
	sbuff = malloc(512);
	v = 1;
//	spiInit();
	fd = _open("COM1",CIO_WRITE_ONLY | CIO_READ_ONLY);
	//----------------------------
	// initialize MMC card
	//----------------------------
	while(1)
	{
		msg = MqGet(SDcardMSG);
		switch(msg->MsgCmd)
		{
			case MMC_CMD_RESET:
				wprintw(win,"MMC Init\n");
				v = mmcReset();
				wprintw(win,"MMC Init Result = %x\n",v);
				break;
			case MMC_CMD_RDSECTOR:
				pL = (unsigned long *)msg->Payload;
				for(i=0;i<512;++i) sbuff[i] = 0xcd;
				v = mmcRead(*pL,sbuff);
				wprintw(win,"Read Sector %ld %02x\n",*pL,v & 0x0ff);
				i = sprintf(b,"-----SECTOR %lu------- %02x\r\n",*pL,v);
				_write(fd,b,i);
				SecDump(fd,(char *)sbuff);
				break;
			case MMC_CMD_WDSECTOR:
				pL = (unsigned long *)msg->Payload;
				pU = (unsigned short *)&msg->Payload[sizeof(unsigned long)];
				wprintw(win,"Write Sector %lu  %u\n",*pL,*pU);
				HexDump(fd,msg->Payload,64);
				memset(sbuff,0,512);
				memcpy(sbuff,&msg->Payload[sizeof(unsigned long)+sizeof(unsigned short)],*pU);
				mmcWrite(*pL, sbuff);
				break;
			case MMC_CMD_STATUS:
				uS = mmcSendStatus();
				wprintw(win,"Status = %04x\n",uS);
				break;
			case MMC_CMD_WRITE_FILE:
				FileObj = (FIL *)malloc(sizeof(FIL));
				FSYS = (FATFS *)malloc(sizeof(FATFS));
				if(f_mount(1,FSYS) != 0) wprintw(win,"Mount Error\n");
				if(f_open(FileObj,"1:test1.txt",FA_WRITE | FA_CREATE_NEW) == 0)
				{
					sprintf((char *)sbuff,"This is a TEST\r\n");
					for(i=0;i<50;++i)
					{
						f_write(FileObj,sbuff,strlen((char *)sbuff),&uI);
					}
					f_sync(FileObj);
					f_close(FileObj);
					wprintw(win,"DONE\n");
				}
				else
				{
					wprintw(win,"Error\n");
				}
				break;
		}
		DeleteMSG(msg);		//return memory used by message
	}
}

void MsgMMCReset(void)
{
	MSG *msg;

	msg = NewMSG(0);
	msg->MsgCmd = MMC_CMD_RESET;
	MqPut(SDcardMSG,msg);
}

void MsgMMCReadSector(unsigned long sector)
{
	MSG *msg;
	unsigned long *pL;

	msg = NewMSG(sizeof(unsigned long));
	msg->MsgCmd = MMC_CMD_RDSECTOR;
	pL = (unsigned long *)msg->Payload;
	*pL = sector;
	MqPut(SDcardMSG,msg);
}

void MsgMMCWriteSector(unsigned long sector,char *b,int n)
{
	MSG *msg;
	unsigned long *pL;
	unsigned short *pU;

	msg = NewMSG(n+sizeof(unsigned long)+sizeof(unsigned short));
	msg->MsgCmd = MMC_CMD_WDSECTOR;
	memcpy(&msg->Payload[sizeof(unsigned long)+sizeof(unsigned short)],b,n );
	pL = (unsigned long *)msg->Payload;
	pU = (unsigned short *)&msg->Payload[sizeof(unsigned long)];
	*pL = sector;
	*pU = n;
	MqPut(SDcardMSG,msg);
}

void MsgMMCSendStatus(void)
{
	MSG *msg;

	msg = NewMSG(0);
	msg->MsgCmd = MMC_CMD_STATUS;
	MqPut(SDcardMSG,msg);
}

void MsgMMCWriteTestFile(void)
{
	MSG *msg;

	msg = NewMSG(0);
	msg->MsgCmd = MMC_CMD_WRITE_FILE;
	MqPut(SDcardMSG,msg);
}


