/**
* Multitasking Fucntions
*/

#ifndef TASK__H
#define TASK__H

#include "pq.h"

/*!
 * \brief MultiTasking Kernel
 * \file task.h
 */

/*!
 * \addtogroup MultitaskingKernel
 */
/*@{*/

#define LO(x)		(x & 0x0ff)			/*!< Low 8 bits of word	*/
#define HI(x)		((x >> 8) & 0x0ff)	/*!< High 8 bits of word	*/

#define STACK_SIZE			96		/*!< Default Stack Size */

#define EVENT_NOERROR		0		/*!< Event Succesful	*/
#define	EVENT_NOTASKS		-1		/*!< No Tasks in Priority Queue */
#define EVENT_OVERFLOW		-2		/*!< Event count overflowed	*/
#define EVENT_TIMEOUT		-3		/*!< Event timed out		*/
#define EVENT_DELETED		-4		/*!< Event was deleted by task */


/*!
 * \struct tcb
 * \brief Definition of a Task Control Block (TCB)
 *
 * \var stack points to the current position of the SP in the stack
 * \var stacktop points to the end of the stack (lowest address) <br>
 * This is where you would place the result of malloc when you allocate stack space
 * \var stacksize size of the stack in bytes
 * \var priority priority level of the task, bit numbers = high priority
 * \var status Status of task, semaphores set this value when they are posted
 * \var misc this value can be set by the user.  It is sometimes useful
 * \var TimeStamp This is the time when the task was swapped out from the timerticker
 * \var TcbSwaps Number of times the task has been swapped out
 * \var timeout Number of timerticks to wait to time out a semaphore waiting on the task
 * \var name task name...this should be unique to aid in debugging
 * \var next pointer to the next TCB, used in pending list of semaphores
 * \var list pointer to master list of TCBs that have been created
 */

typedef struct tcb TCB;

struct tcb {
	char *stack;		/*!<	points to current position in stack	*/
	char *stacktop;		/*!<	points to the end of the stack		*/
	int stacksize;		/*!<	SIZE of the stack in bytes			*/
	int priority;		/*!<	task priority	*/
	int status;			/*!<	task status, set by semaphore mostly	*/
	unsigned misc;		/*!<	misc data user defined		*/
	unsigned TimeStamp;		/*!<	system time stamp, set when swapped	*/
	unsigned TcbSwaps;		/*!<	number of times tasks swapped in	*/
	int timeout;        /*!<	timeout counter to trigger event	*/
	char *name;			/*!<	name of task	*/
	TCB *next;
	TCB *list;	/*!<	master linked list of tasks	*/
};

/*!
 * Semaphore Object
 */
typedef struct ecb {
	TCB *task_h;		/*!< head pointer for list of waiting tasks	*/
	TCB *task_t;		/*!< tail pointer for list of waiting tasks	*/
	struct ecb *Tnext,*Tprev;	/*!< Link list pointers for timeout list	*/
	int SemaphoreMode;		/*!< mode of semaphore	*/
	int EventCount;			/*!< counter for semaphore	*/
	char name[20];			/*!< name of Event (Semaphore)	*/
} ECB;						/*!< Event Control Block	*/

#define SEMAPHORE_MODE_BLOCKING		0	/*!<	Semaphore in Blocking Mode	*/
#define SEMAPHORE_MODE_TIMEOUT		1	/*!<	Semaphore in Timeout Mode	*/

//-------------------------------------------------------------------
// Thread Definition macro
//-------------------------------------------------------------------

#define THREAD(func,arg)\
void func(void *arg) __attribute__((noreturn)); \
void func(void *arg)	/*!<	Macro used to define a thread function	*/
	

/*-------------------------------------------------------------------
** Function Prototypes
**------------------------------------------------------------------*/

extern ECB *MasterList;	//head pointer for master semaphore list
extern ECB *TimeoutList;	//head pointer for timeout semaphore list
extern PQ ActiveTasks;		//priority task queue
extern TCB *CurrentTask,*NextTask;
extern ECB *LCDBlocker;
extern int OsRunning;
extern volatile int TStamp;

extern volatile int InterruptCount;
extern volatile int Blocking;

extern void CreateTask(TCB *t, void (*task)(void *), int stacksize,int priority,char *name,void *arg);
extern void IrqSwap(void);
extern void OSInit(void);
extern void Yeild(void);
extern ECB *NewSemaphore(int InitCount,int Mode,char *name);
extern void CreateSemaphore(ECB *e,int InitCount,int Mode,char *n);
extern void DeleteSemaphore(ECB *e);
extern int PendSemaphore(ECB *e, int Timeout);	//wait for semaphore availiable
extern int PostSemaphore(ECB *e, int Value);	//signal semaphore availiable
extern int PostSemaphoreWithData(ECB *e, int Value,unsigned data);	//signal semaphore availiable
extern int PostSemaphoreIrq(ECB *e, int Value);	//signal semaphore availiable
extern void TimerTicker(void);
extern void vPortYieldProcessor(void);
extern void vPortStartFirstTask(void);
extern int TimeDelay(int mSec);
extern void Enable(char sr);
extern char Disable(void);
extern void StartOS(void);
extern void DoSwap(void);
extern void ExitInterrupt(void);
extern unsigned MeasureStackUsage(char *s,unsigned max);
extern unsigned StackHeapAvailiable(void);
extern void RegisterTickerPost(ECB *pS);

//----------------------------------------------------------------
// Macros that are used in interrupt routines for saving
// and restoring the context
// it should be noted in the save IRQ sequence that we
// set the interrupt enable bit in the value we get from
// the status register.  If we got here, interrupts were
// enabled, so we want to make sure that is how we
// exit.
//----------------------------------------------------------------

/*!
 * \def SAVE_IRQ()
 * \brief Save context in Interrupt Routine
 *
 * this macro is used to save processor context withing an
 * interrupt routine
 */

#define SAVE_IRQ()									\
	asm volatile (	"push	r31						\n\t"	\
					"in		r31, __SREG__			\n\t"	\
					"ori	r31,0x80					\n\t"	\
					"push	r31						\n\t"	\
					"push	r30						\n\t"	\
					"push	r29						\n\t"	\
					"push	r28						\n\t"	\
					"push	r27						\n\t"	\
					"push	r26						\n\t"	\
					"push	r25						\n\t"	\
					"push	r24						\n\t"	\
					"push	r23						\n\t"	\
					"push	r22						\n\t"	\
					"push	r21						\n\t"	\
					"push	r20						\n\t"	\
					"push	r19						\n\t"	\
					"push	r18						\n\t"	\
					"push	r17						\n\t"	\
					"push	r16						\n\t"	\
					"push	r15						\n\t"	\
					"push	r14						\n\t"	\
					"push	r13						\n\t"	\
					"push	r12						\n\t"	\
					"push	r11						\n\t"	\
					"push	r10						\n\t"	\
					"push	r9						\n\t"	\
					"push	r8						\n\t"	\
					"push	r7						\n\t"	\
					"push	r6						\n\t"	\
					"push	r5						\n\t"	\
					"push	r4						\n\t"	\
					"push	r3						\n\t"	\
					"push	r2						\n\t"	\
					"push	r1						\n\t"	\
					"clr	r1						\n\t"	\
					"push	r0						\n\t"	\
				);

#define RESTORE_IRQ()								\
	asm volatile (	"pop	r0						\n\t"	\
					"pop	r1						\n\t"	\
					"pop	r2						\n\t"	\
					"pop	r3						\n\t"	\
					"pop	r4						\n\t"	\
					"pop	r5						\n\t"	\
					"pop	r6						\n\t"	\
					"pop	r7						\n\t"	\
					"pop	r8						\n\t"	\
					"pop	r9						\n\t"	\
					"pop	r10						\n\t"	\
					"pop	r11						\n\t"	\
					"pop	r12						\n\t"	\
					"pop	r13						\n\t"	\
					"pop	r14						\n\t"	\
					"pop	r15						\n\t"	\
					"pop	r16						\n\t"	\
					"pop	r17						\n\t"	\
					"pop	r18						\n\t"	\
					"pop	r19						\n\t"	\
					"pop	r20						\n\t"	\
					"pop	r21						\n\t"	\
					"pop	r22						\n\t"	\
					"pop	r23						\n\t"	\
					"pop	r24						\n\t"	\
					"pop	r25						\n\t"	\
					"pop	r26						\n\t"	\
					"pop	r27						\n\t"	\
					"pop	r28						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r31						\n\t"	\
					"out	__SREG__, r31			\n\t"	\
					"pop	r31						\n\t"	\
					"reti							\n\t"	\
				);
//-------------------------------------------------------------------
// macros used for saving and restoring context from a task level
//-------------------------------------------------------------------

#define SAVE_CONTEXT()									\
	asm volatile (	"push	r31						\n\t"	\
					"in		r31, __SREG__			\n\t"	\
					"cli							\n\t"	\
					"push	r31						\n\t"	\
					"push	r30						\n\t"	\
					"push	r29						\n\t"	\
					"push	r28						\n\t"	\
					"push	r27						\n\t"	\
					"push	r26						\n\t"	\
					"push	r25						\n\t"	\
					"push	r24						\n\t"	\
					"push	r23						\n\t"	\
					"push	r22						\n\t"	\
					"push	r21						\n\t"	\
					"push	r20						\n\t"	\
					"push	r19						\n\t"	\
					"push	r18						\n\t"	\
					"push	r17						\n\t"	\
					"push	r16						\n\t"	\
					"push	r15						\n\t"	\
					"push	r14						\n\t"	\
					"push	r13						\n\t"	\
					"push	r12						\n\t"	\
					"push	r11						\n\t"	\
					"push	r10						\n\t"	\
					"push	r9						\n\t"	\
					"push	r8						\n\t"	\
					"push	r7						\n\t"	\
					"push	r6						\n\t"	\
					"push	r5						\n\t"	\
					"push	r4						\n\t"	\
					"push	r3						\n\t"	\
					"push	r2						\n\t"	\
					"push	r1						\n\t"	\
					"clr	r1						\n\t"	\
					"push	r0						\n\t"	\
					"lds	r26, CurrentTask		\n\t"	\
					"lds	r27, CurrentTask + 1	\n\t"	\
					"in		r0, 0x3d				\n\t"	\
					"st		x+, r0					\n\t"	\
					"in		r0, 0x3e				\n\t"	\
					"st		x+, r0					\n\t"	\
				);

/* 
 * Opposite to SAVE_CONTEXT().  Interrupts will have been disabled during
 * the context save so we can write to the stack pointer. 
 */
/*
#define RESTORE_CONTEXT()								\
	asm volatile (	"lds	r26, NextTask		\n\t"	\
					"lds	r27, NextTask + 1	\n\t"	\
					"ld		r28, x+					\n\t"	\
					"out	__SP_L__, r28			\n\t"	\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r0						\n\t"	\
					"pop	r1						\n\t"	\
					"pop	r2						\n\t"	\
					"pop	r3						\n\t"	\
					"pop	r4						\n\t"	\
					"pop	r5						\n\t"	\
					"pop	r6						\n\t"	\
					"pop	r7						\n\t"	\
					"pop	r8						\n\t"	\
					"pop	r9						\n\t"	\
					"pop	r10						\n\t"	\
					"pop	r11						\n\t"	\
					"pop	r12						\n\t"	\
					"pop	r13						\n\t"	\
					"pop	r14						\n\t"	\
					"pop	r15						\n\t"	\
					"pop	r16						\n\t"	\
					"pop	r17						\n\t"	\
					"pop	r18						\n\t"	\
					"pop	r19						\n\t"	\
					"pop	r20						\n\t"	\
					"pop	r21						\n\t"	\
					"pop	r22						\n\t"	\
					"pop	r23						\n\t"	\
					"pop	r24						\n\t"	\
					"pop	r25						\n\t"	\
					"pop	r26						\n\t"	\
					"pop	r27						\n\t"	\
					"pop	r28						\n\t"	\
					"pop	r29						\n\t"	\
					"pop	r30						\n\t"	\
					"pop	r31						\n\t"	\
					"out	__SREG__, r31			\n\t"	\
					"pop	r31						\n\t"	\
				);
*/

#define RESTORE_CONTEXT_OLD()								\
	asm volatile (	"lds	r26, NextTask		\n\t"	/* Get Address of Next Task TCB*/ \
					"lds	r27, NextTask + 1	\n\t"	\
					"ld		r28, x+					\n\t"	/* load pointer to Stack Pointer */\
					"out	__SP_L__, r28			\n\t"	/* load Stack pointer for restored Task */\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r0						\n\t"	/* Pop R0 from Stack */ \
					"pop	r1						\n\t"	/* Pop R1 from Stack */ \
					"pop	r2						\n\t"	/* Pop R2 from Stack */ \
					"pop	r3						\n\t"	/* Pop R3 from Stack */ \
					"pop	r4						\n\t"	/* Pop R4 from Stack */ \
					"pop	r5						\n\t"	/* Pop R5 from Stack */ \
					"pop	r6						\n\t"	/* Pop R6 from Stack */ \
					"pop	r7						\n\t"	/* Pop R7 from Stack */ \
					"pop	r8						\n\t"	/* Pop R8 from Stack */ \
					"pop	r9						\n\t"	/* Pop R9 from Stack */ \
					"pop	r10						\n\t"	/* Pop R10 from Stack */ \
					"pop	r11						\n\t"	/* Pop R11 from Stack */ \
					"pop	r12						\n\t"	/* Pop R12 from Stack */ \
					"pop	r13						\n\t"	/* Pop R13 from Stack */ \
					"pop	r14						\n\t"	/* Pop R14 from Stack */ \
					"pop	r15						\n\t"	/* Pop R15 from Stack */ \
					"pop	r16						\n\t"	/* Pop R16 from Stack */ \
					"pop	r17						\n\t"	/* Pop R17 from Stack */ \
					"pop	r18						\n\t"	/* Pop R18 from Stack */ \
					"pop	r19						\n\t"	/* Pop R19 from Stack */ \
					"pop	r20						\n\t"	/* Pop R20 from Stack */ \
					"pop	r21						\n\t"	/* Pop R21 from Stack */ \
					"pop	r22						\n\t"	/* Pop R22 from Stack */ \
					"pop	r23						\n\t"	/* Pop R23 from Stack */ \
					"pop	r24						\n\t"	/* Pop R24 from Stack */ \
					"pop	r25						\n\t"	/* Pop R25 from Stack */ \
					"pop	r26						\n\t"	/* Pop R26 from Stack */ \
					"pop	r27						\n\t"	/* Pop R27 from Stack */ \
					"pop	r28						\n\t"	/* Pop R28 from Stack */ \
					"pop	r29						\n\t"	/* Pop R29 from Stack */ \
					"pop	r30						\n\t"	/* Pop R30 from Stack */ \
					"pop	r31						\n\t"	/* Pop Status from Stack */ \
					"tst	r31						\n\t"	/* Check IRQ en bit */ \
					"brmi	irqen					\n\t"	/* Is Irq Enabled? */ \
					"out	__SREG__, r31			\n\t"	/* No, update Status Reg */ \
					"pop	r31						\n\t"	/* Pop R31 from Stack */ \
					"ret							\n\t"	/* Do normal return */ \
					"irqen: cbr 	r31, 128		\n\t"	/* YES, clear IRQ En Bit */ \
					"out	__SREG__, r31			\n\t"	/* Update Status Reg */ \
					"pop	r31						\n\t"	/* Pop R31 from Stack */ \
					"reti							\n\t"	/* return from IRQ (enable IRQ on Return */ \
				);


#define RESTORE_CONTEXT()								\
	asm volatile (	"lds	r26, NextTask		\n\t"	/* Get Address of Next Task TCB*/ \
					"lds	r27, NextTask + 1	\n\t"	\
					"ld		r28, x+					\n\t"	/* load pointer to Stack Pointer */\
					"out	__SP_L__, r28			\n\t"	/* load Stack pointer for restored Task */\
					"ld		r29, x+					\n\t"	\
					"out	__SP_H__, r29			\n\t"	\
					"pop	r0						\n\t"	/* Pop R0 from Stack */ \
					"pop	r1						\n\t"	/* Pop R1 from Stack */ \
					"pop	r2						\n\t"	/* Pop R2 from Stack */ \
					"pop	r3						\n\t"	/* Pop R3 from Stack */ \
					"pop	r4						\n\t"	/* Pop R4 from Stack */ \
					"pop	r5						\n\t"	/* Pop R5 from Stack */ \
					"pop	r6						\n\t"	/* Pop R6 from Stack */ \
					"pop	r7						\n\t"	/* Pop R7 from Stack */ \
					"pop	r8						\n\t"	/* Pop R8 from Stack */ \
					"pop	r9						\n\t"	/* Pop R9 from Stack */ \
					"pop	r10						\n\t"	/* Pop R10 from Stack */ \
					"pop	r11						\n\t"	/* Pop R11 from Stack */ \
					"pop	r12						\n\t"	/* Pop R12 from Stack */ \
					"pop	r13						\n\t"	/* Pop R13 from Stack */ \
					"pop	r14						\n\t"	/* Pop R14 from Stack */ \
					"pop	r15						\n\t"	/* Pop R15 from Stack */ \
					"pop	r16						\n\t"	/* Pop R16 from Stack */ \
					"pop	r17						\n\t"	/* Pop R17 from Stack */ \
					"pop	r18						\n\t"	/* Pop R18 from Stack */ \
					"pop	r19						\n\t"	/* Pop R19 from Stack */ \
					"pop	r20						\n\t"	/* Pop R20 from Stack */ \
					"pop	r21						\n\t"	/* Pop R21 from Stack */ \
					"pop	r22						\n\t"	/* Pop R22 from Stack */ \
					"pop	r23						\n\t"	/* Pop R23 from Stack */ \
					"pop	r24						\n\t"	/* Pop R24 from Stack */ \
					"pop	r25						\n\t"	/* Pop R25 from Stack */ \
					"pop	r26						\n\t"	/* Pop R26 from Stack */ \
					"pop	r27						\n\t"	/* Pop R27 from Stack */ \
					"pop	r28						\n\t"	/* Pop R28 from Stack */ \
					"pop	r29						\n\t"	/* Pop R29 from Stack */ \
					"pop	r30						\n\t"	/* Pop R30 from Stack */ \
					"pop	r31						\n\t"	/* Pop Status from Stack */ \
					"sbrc	r31,7					\n\t"	/* Check IRQ en bit */ \
					"rjmp	irqen					\n\t"	/* Is Irq Enabled? */ \
					"out	__SREG__, r31			\n\t"	/* No, update Status Reg */ \
					"pop	r31						\n\t"	/* Pop R31 from Stack */ \
					"ret							\n\t"	/* Do normal return */ \
					"irqen: cbr 	r31, 128		\n\t"	/* YES, clear IRQ En Bit */ \
					"out	__SREG__, r31			\n\t"	/* Update Status Reg */ \
					"pop	r31						\n\t"	/* Pop R31 from Stack */ \
					"reti							\n\t"	/* return from IRQ (enable IRQ on Return */ \
				);

/*@}*/

/*
**
*/

#endif
