/****************************************************************
** This file contains the objects for creating an event trigger
**
** It creates an event based on a trigger variable, and the past
** history of that variable.
**
** The event is signaled by calling a callback function
**
** Created May 7, 2008
**
****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "EventTrigger.h"

EVTB **EventsList;


EVTB *NewEventTrigger(unsigned char mode, void (*cb)(char,void*),void *arg)
{
	//------------------------------------------------------------------------
	// Create an Event Trigger Block
	//
	// parameters:
	//	mode.....trigger mode (positive (low->high), negative (high->low), both
	//	cb.......pointer to callback function to execute on event true
	//	arg......pointer to arguments to pass to callback function
	//
	// return value:
	//		returns pointer to EVTB on success, NULL on fail
	//------------------------------------------------------------------------
	EVTB *rv;

	rv = (EVTB *) malloc(sizeof(EVTB));
	if(rv)
	{
		rv->Previous = 0;
		rv->TrigMode = mode;
		rv->callback = cb;
		rv->arg = arg;
	}
	return rv;
}

void FreeEventTrigger(EVTB *pE)
{
	//-------------------------------------------------------------------------
	// Free an Event Trigger Block
	//
	// parameter:
	//	pE......pointer to EVTB to free
	//-------------------------------------------------------------------------
	free((void *)pE);
}

int EventTrigger(EVTB *pE,unsigned char v)
{
	//----------------------------------------------------------------------------
	// Process an event....
	// v is the value from the event.  If v is greater than the preveous value
	// then generate a postive trigger event if trigger mode is set to postive.
	// if v is less than the previous value and the trigger mode is set to negative
	// then generate a negative trigger event.
	// We can generate an event on either positive or negative transistions, or
	// just positive or just negative.
	//
	// parameters:
	//	pE..........pointer to the event control block
	//	v...........value that controls the event
	//----------------------------------------------------------------------------
	int rv = 0;
	if(v != pE->Previous)
	{
		if(( ( (v - pE->Previous) > 0) ) && (pE->TrigMode & EVENTTRIGGER_MODE_POS) )
		{
			if(pE->callback) (*pE->callback)(EVENTTRIGGER_MODE_POS,pE->arg);
			rv = 1;
		}
		else if (( ( (v - pE->Previous) < 0) ) && pE->TrigMode & EVENTTRIGGER_MODE_NEG)
		{
			if(pE->callback) (*pE->callback)(EVENTTRIGGER_MODE_NEG,pE->arg);
			rv = 1;
		}
		pE->Previous = v;
	}
	return rv;
}

void InitEventTriggers(int n)
{
	int i;

	EventsList = malloc(sizeof(EVTB *) * n);
	for(i=0;i<n;++i)
		EventsList[i] = NewEventTrigger(EVENTTRIGGER_MODE_POS, (void (*)(char,void*))NULL,NULL);
}

int CheckEventTrigger(int e,int v)
{
	return EventTrigger(EventsList[e],v);
}

