/*------------------------------------------------------------------------
 * This file is used to manage heaps
 *
 * Created April 2, 2008 by Jim Patchell
 *
 * Now, this is derived from malloc and free that comes with WinAVR and
 * following is the copyright notice that was in THAT code.  It is mostly
 * unmodified.
 * -----------------------------------------------
Copyright (c) 2002, 2004 Joerg Wunsch
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.

   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.

   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
* -----------------------------------------------------------------------
* Code first modified on April 2, 2008
---------------------------------------------------------------------------*/

#include <stdlib.h>
#include "HeapManager.h"
#include <avr/io.h>

/*!
 * \file heapmanager.c
 * \ingroup MultitaskingKernel
 */

/*@{*/

/*!
 * \fn void *HeapAlloc(HEAP_BLOCK *pHB,size_t len)
 * \brief This function is used to allocate memory from a heap
 *
 * The code for this function was lifted directly out of the WINAVR
 * libc library.  This was originally malloc writen by Joerg Wunsch
 * and modified by Jim Patchell to turn it into a bit more of an
 * object oriented function.  The object is a HEAP_BLOCK.
 *
 * \param pHB pointer to a heap block object
 * \param len size of desired memory block
 * \return pointer to memory block, or NULL on FAIL!
 */

void *HeapAlloc(HEAP_BLOCK *pHB,size_t len)
{
	struct __freelist *fp1, *fp2;
	char *cp;
	size_t s, avail;

	/*
	 * Our minimum chunk size is the size of a pointer (plus the
	 * size of the "sz" field, but we don't need to account for
	 * this), otherwise we could not possibly fit a freelist entry
	 * into the chunk later.
	 */
	PendSemaphore(pHB->Blocker,0);
	if (len < sizeof(struct __freelist) - sizeof(size_t))
		len = sizeof(struct __freelist) - sizeof(size_t);

	/*
	 * First, walk the free list and try finding a chunk that
	 * would match exactly.  If we found one, we are done.  While
	 * walking, note down the size of the largest chunk we found
	 * that would still fit the request -- we need it for step 2.
	 *
	 */
	for (s = 0, fp1 = pHB->__flp, fp2 = 0;
	     fp1;
	     fp2 = fp1, fp1 = fp1->nx) {
		if (fp1->sz == len) {
			/*
			 * Found it.  Disconnect the chunk from the
			 * freelist, and return it.
			 */
			if (fp2)
				fp2->nx = fp1->nx;
			else
				pHB->__flp = fp1->nx;
			PostSemaphore(pHB->Blocker,0);
			return &(fp1->nx);
		}
		if (fp1->sz > len) {
			if (s == 0 || fp1->sz < s)
				s = fp1->sz;
		}
	}
	/*
	 * Step 2: If we found a chunk on the freelist that would fit
	 * (but was too large), look it up again and use it, since it
	 * is our closest match now.  Since the freelist entry needs
	 * to be split into two entries then, watch out that the
	 * difference between the requested size and the size of the
	 * chunk found is large enough for another freelist entry; if
	 * not, just enlarge the request size to what we have found,
	 * and use the entire chunk.
	 */
	if (s) {
		if (s - len < sizeof(struct __freelist))
			len = s;
		for (fp1 = pHB->__flp, fp2 = 0;
		     fp1;
		     fp2 = fp1, fp1 = fp1->nx) {
			if (fp1->sz == s) {
				if (len == s) {
					/*
					 * Use entire chunk; same as
					 * above.
					 */
					if (fp2)
						fp2->nx = fp1->nx;
					else
						pHB->__flp = fp1->nx;
					PostSemaphore(pHB->Blocker,0);
					return &(fp1->nx);
				}
				/*
				 * Split them up.  Note that we leave
				 * the first part as the new (smaller)
				 * freelist entry, and return the
				 * upper portion to the caller.  This
				 * saves us the work to fix up the
				 * freelist chain; we just need to
				 * fixup the size of the current
				 * entry, and note down the size of
				 * the new chunk before returning it
				 * to the caller.
				 */
				cp = (char *)fp1;
				s -= len;
				cp += s;
				fp2 = (struct __freelist *)cp;
				fp2->sz = len;
				fp1->sz = s - sizeof(size_t);
				PostSemaphore(pHB->Blocker,0);
				return &(fp2->nx);
			}
		}
	}
	/*
	 * Step 3: If the request could not be satisfied from a
	 * freelist entry, just prepare a new chunk.  This means we
	 * need to obtain more memory first.  The largest address just
	 * not allocated so far is remembered in the brkval variable.
	 * Under Unix, the "break value" was the end of the data
	 * segment as dynamically requested from the operating system.
	 * Since we don't have an operating system, just make sure
	 * that we don't collide with the stack.
	 */
	if (pHB->BrkVal == 0)
		pHB->BrkVal = pHB->Start;
	cp = pHB->End;
	if (cp == 0)
		cp = STACK_POINTER() - pHB->Margin;
	avail = cp - pHB->BrkVal;
	/*
	 * Both tests below are needed to catch the case len >= 0xfffe.
	 */
	if (avail >= len && avail >= len + sizeof(size_t)) {
		fp1 = (struct __freelist *)pHB->BrkVal;
		pHB->BrkVal += len + sizeof(size_t);
		fp1->sz = len;
		PostSemaphore(pHB->Blocker,0);
		return &(fp1->nx);
	}
	/*
	 * Step 4: There's no help, just fail. :-/
	 */
	 PostSemaphore(pHB->Blocker,0);
	return 0;
}

/*!
 * \fn void HeapFree(HEAP_BLOCK *pHB,void *p)
 * \brief Free a block of memory back to a heap
 *
 * The code for this function was lifted directly out of the WINAVR
 * libc library.  This was originally malloc writen by Joerg Wunsch
 * and modified by Jim Patchell to turn it into a bit more of an
 * object oriented function.  The object is a HEAP_BLOCK.
 * 
 * \param pHB pointer to a HEAP_BLOCK object to free memory to
 * \param p pointer to the block of memory to free
 */

void HeapFree(HEAP_BLOCK *pHB,void *p)
{
	struct __freelist *fp1, *fp2, *fpnew;
	char *cp1, *cp2, *cpnew;

	/* ISO C says free(NULL) must be a no-op */
	if (p == 0)
		return;

	PendSemaphore(pHB->Blocker,0);
	cpnew = p;
	cpnew -= sizeof(size_t);
	fpnew = (struct __freelist *)cpnew;
	fpnew->nx = 0;

	/*
	 * Trivial case first: if there's no freelist yet, our entry
	 * will be the only one on it.
	 */
	if (pHB->__flp == 0) {
		pHB->__flp = fpnew;
		PostSemaphore(pHB->Blocker,0);
		return;
	}

	/*
	 * Now, find the position where our new entry belongs onto the
	 * freelist.  Try to aggregate the chunk with adjacent chunks
	 * if possible.
	 */
	for (fp1 = pHB->__flp, fp2 = 0;
	     fp1;
	     fp2 = fp1, fp1 = fp1->nx) {
		if (fp1 < fpnew)
			continue;
		cp1 = (char *)fp1;
		fpnew->nx = fp1;
		if ((char *)&(fpnew->nx) + fpnew->sz == cp1) {
			/* upper chunk adjacent, assimilate it */
			fpnew->sz += fp1->sz + sizeof(size_t);
			fpnew->nx = fp1->nx;
		}
		if (fp2 == 0) {
			/* new head of freelist */
			pHB->__flp = fpnew;
			PostSemaphore(pHB->Blocker,0);
			return;
		}
		break;
	}
	/*
	 * Note that we get here either if we hit the "break" above,
	 * or if we fell off the end of the loop.  The latter means
	 * we've got a new topmost chunk.  Either way, try aggregating
	 * with the lower chunk if possible.
	 */
	fp2->nx = fpnew;
	cp2 = (char *)&(fp2->nx);
	if (cp2 + fp2->sz == cpnew) {
		/* lower junk adjacent, merge */
		fp2->sz += fpnew->sz + sizeof(size_t);
		fp2->nx = fpnew->nx;
	}
	PostSemaphore(pHB->Blocker,0);
}

//-----------------------------------------------------------------------
// The following code was added by Jim Patchell on April 2, 2008
//-----------------------------------------------------------------------

/*!
 * \fn HEAP_BLOCK * HeapInit(char *start, char *end)
 * \brief Initialize a heap block object
 *
 * this function is used to initialize a heap block.  The
 * user passes this function a pointer to the start and end
 * if a block of memory that is going to be used as the heap.
 * This function does a lot of things to that memory.  One is
 * to create a semaphore in the block of memory so that calling
 * HeapAlloc and HeapFree will be thread safe.
 *
 * So, one must remember that not all of the block will be
 * available for allocation.
 *
 * \param start pointer to start of memory to use for heap
 * \param end pointer to end of memory to use for heap
 * \return pointer to HEAP_BLOCK object upon success
 */

HEAP_BLOCK * HeapInit(char *start, char *end)
{
	//--------------------------------------------------------------------
	// Initialize semaphore used to block access to malloc and free
	// parameters:
	//	start.........start of heap
	//	end...........end of heap
	//
	// We use memory in the defined heap to allocate memory for the
	// object that is used to manage the memory
	//
	// return value:
	//	pointer to heap object
	//--------------------------------------------------------------------
	HEAP_BLOCK *pHB = (HEAP_BLOCK *)start;
	start += sizeof (HEAP_BLOCK);	//allocate heap control block
	pHB->Blocker = (ECB *)start;
	start += sizeof (ECB);			//allocate event control block
	pHB->Start = start;
	pHB->End = end;
	pHB->BrkVal = start;
	pHB->Margin = 32;		//default value for stack margin...not needed..really
	pHB->__flp = 0;			//free list is empty
	CreateSemaphore(pHB->Blocker,1,SEMAPHORE_MODE_BLOCKING,"Malloc");
	return pHB;
}

/*@}*/


