`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    08:10:43 02/07/2007 
// Design Name: 
// Module Name:    muart 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: Uart Core
//
// Using the UART Core:
// inputs:
//	clk..........system clock input
// reset........system reset (active high)
// D............8 bit data bus input to uart
// O............8 bit data bus output from uart (read registers back)
// A............3 bit address select bus
// wd...........write data to registers
// rd...........read data back (a strobe actualy...most registers don't need this
// I............Interrupt signals
// rxd..........serial input into uart
// txd..........serial output from uart
//
//---------------------------------------------
// Register Definitions:
//
// A == 0
// Data Register
//		Reads and writes the receive and transmit fifos.  When data is written into
// the transmit fifo, it is moved into the transmit shift register as soon as
// posible.  The receive fifo uses the data strobe to read the first data word
// into the fifo output register.  This means that the read stobe must complete
// its cycle (needs to go high for one clock cycle) before the first byte can be
// read.
//-----------------------------------------------
// A == 1
// receive Uart Status Register
//		This register indicates the current status of the receive uart
//	BIT0-3:Byte Count in FIFO
//	BIT4: Fifo FULL
// BIT5: Fifo Empty
// BIT6: Overrun Error (not implemented yet 2-12-2007)
// BIT7: Frame Error (not implemented yet 2-12-2007)
//
//----------------------------------------------
// A == 2
// Transmit Uart Status Register
//		This register indicates the current status of the transmit uart
// BIT0-3:Byte Count in FIFO
// BIT4: Fifo FULL
// BIT5: Fifo EMPTY
// BIT6: reads 0 (reserved)
// BIT7: reads 0 (reserved)
//----------------------------------------------
// A == 3 (Baud Rate Low Byte)
// A == 4 (Baud Rate High Byte)
// Baud Rate Register
//
// This register sets the baud rate for the uart.
//
// Baud rate is calculated as follows:
//
// Baudrate_Register = ((Baudrate_in_BPS)/(Master_Clock_Rate)) * *(2^16) * 16);
//
// Example: Master Clock is 50MHz, Baud Rate needs to be 19200
//
// 19200/50000000*2^16 * 16 = 403. (.086% error)
//
// It should be noted that the maximum value that can be written into
// the Baud Rate Generator is 0x7fff (32767).  A value of 0x4000 (16384) at 50MHz
// will give a baud rate of 781250 bps...which I consider the highest practical 
// baud rate
//
// A value greater than 32767 will give very strange results, indeed...
//
// The baudrate generator is implemented using a accumulator divider (NCO)
// and unlike a divider, the clock frequency is not quite as critical.
//
// For example, if the clock rate is 50MHz, a 115200bps baud rate will
// have a 0.469% error.
//
// with the NCO baud rate generator, the error is 0.003%.
//
// Although, at the slower baud rates, you will tend to have higher
// errors.
//
//-------------------------------------------------
// A == 5
// Interrupt Enable Register
//
// Each bit set high will enable the interrupt specified in the 
// Interrupt Flag Register (see below)
//
// BIT0:Receive Fifo has characters waiting
// BIT1:Transmit Fifo is Empty
// BIT2:Tranmist Uart has sent last character
// BIT3-7:(reserved)
//
//-------------------------------------------------
// A == 6
// Interrupt Flag Register
//
// These bits indicate the current status of its
// corresponding interrupt
// If the interrupt enable bit is not set to one
// for the coresponding bit in the flag register, that
// bit will read 0 (since it can not cause an interrupt)
// Raw status bits can be read in the transmit and receive
// status registers (see above).
//
// BIT0:Receive Fifo has characters waiting
// BIT1:Transmit Fifo is Empty
// BIT2:Tranmist Uart has sent last character
// BIT3-7:(reserved)
//-------------------------------------------------
// A == 7
// Interrupt Vector Register
//
//		The value in this register indicates which
//	interrupt wants to be serviced
//
//-------------------------------------------------
//
// Interrupts:
//
//		Receive Fifo has characters waiting
//
//	This interrupt is activated anytime there are
// any characters waiting in the receive FIFO.  This
// interrupt can only be cleared by reading all of
// the characters out of the FIFO. (i.e. reading
// the uart data register (see above)).  This interrupt
// is on signal I[0].
//
//		Transmit Fifo is Empty
//
//	This interrupt is activaged anytime there are no characters
// waiting to be transmitted in the transmit FIFO.  This
// interrupt can only be cleared by writing data into
// the transmit FIFO (i.e. write to the uart data register (see above)).
//	This interrupt is on signal I[1].
//
//		Tranmist Uart has sent last character
//
//	This interrupt is activated when the last bit has been sent
// from the transmit shift register AND the transmit FIFO is
// also empty.  This can be used for firing off an event in
// a real time operating system to let the OS know that the
// transmit uart is no longer busy.  This interrupt can be
// cleared two different ways.
//
// Methode 1:write data to the transmit FIFO
// Methode 2:Write to the IFR with BIT2 Set
//********************************************************************************
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module muart(clk, reset, D, O, A, wd, rd, I, rxd, txd);
    input clk;				//system clock
    input reset;			//system reset
    input [7:0] D;		//data bus in
    output [7:0] O;		//data bus out
    input [2:0] A;		//register select
    input wd;				//write data strobe
    input rd;				//read data strobe
    output I;		//interrupt
	input rxd;				//uart input pin
	output txd;				//uart output pin
	
	reg [7:0] O;
	reg I;
	wire txd;
	wire rxd;
	
parameter UART_DATA			=		3'b000;		//data IO address
parameter UART_RSTATUS 		=		3'b001;		//receive status
parameter UART_TSTATUS		=		3'b010;		//transmit status
parameter UART_BRATE_L		=		3'b011;		//Baud Rate Low Byte
parameter UART_BRATE_H		=		3'b100;		//Baud Rate High Byte
parameter UART_IER			=		3'b101;		//interrupt enable register
parameter UART_IFR			=		3'b110;		//interrupt flag register
parameter UART_IVR			=		3'b111;		//interrupt vector register

//-------------------------------------
// Interrupt Bits
//-------------------------------------

reg IRFifoHasChar;
reg ITFifoMT;
reg ITxdDone;

reg [7:0]Ivr;

//-------------------------------------
// receive fifo variables
//-------------------------------------

wire [3:0] RDCount;
wire wdRFifo;
wire [7:0] RFifoO;
wire rdRFifo;
wire REmpty,RFull;

//-------------------------------------
// Transmit fifo variables
//-------------------------------------

wire [3:0] TDCount;
wire [7:0] TFifoO;
reg wdTXD;
wire TEmpty,TFull;	//transmit fifo empty status
wire rdTFifo;	//read data from T fifo

//-------------------------------------
// Baud Rate Register Variables
//-------------------------------------

reg [7:0] BaudL,BaudH;	//baud rate registers
reg wdBL,wdBH;	//write strobes for baud rate register
wire bclk;	//bit rate clock

//--------------------------------------
// receive Uart Variables
//--------------------------------------

wire [7:0] RDout;
wire frame,overrun;

//---------------------------------------
// Transmit Uart Variables
//---------------------------------------

wire wdTxUart;	//write data into transmit uart
wire tbrate;	//transmit uart bit rate
wire done;		//status bit indicates transmit of char is done

//---------------------------------------
// State machine defines
//---------------------------------------
parameter TUART_STATE_IDLE = 2'b00;
parameter TUART_STATE_RDFIFO = 2'b01;
parameter TUART_STATE_WDUART = 2'b10;
parameter TUART_STATE_WAIT = 2'b11;

(* FSM_ENCODING="SEQUENTIAL", SAFE_IMPLEMENTATION="NO" *) reg [1:0] TState = TUART_STATE_IDLE; //state machine for transmit uart

//---------------------------------------
// Interrupt enable register variables
//---------------------------------------

reg [7:0] IER;
reg wdIER;

//---------------------------------------
// Interrupt Flag Register variables
//---------------------------------------

wire [7:0] IFR;
wire wdClrIFR;

parameter IFR_RXD     = 0;
parameter IFR_TXDMT   = 1;
parameter IFR_TXDDONE = 2;

//**********************************************************************
// Methods Follow
//**********************************************************************

//-------------------------------------
// Interrupts
//-------------------------------------

always @ (posedge clk) begin
	//-------------------------------
	// If we read fifo and the fifo
	// is empty, then we clear
	// the interrupt
	//-------------------------------
	IRFifoHasChar <= ~REmpty;
end

always @ (posedge clk) ITFifoMT <= TEmpty;

always @ (posedge clk or posedge reset) begin
	if(reset)
		ITxdDone <= 0;		//init irq
	else if (TEmpty && done)
		ITxdDone <= 1;		//set IRQ
	else if (wdTXD)
		ITxdDone <= 0;		//clear IRQ
	else if (wdClrIFR && D[IFR_TXDDONE])
		ITxdDone <= 0;		//clear IRQ
end

assign IFR[IFR_RXD] = IER[IFR_RXD] & IRFifoHasChar;
assign IFR[IFR_TXDMT] = IER[IFR_TXDMT] & ITFifoMT;
assign IFR[IFR_TXDDONE] = IER[IFR_TXDDONE] & ITxdDone;
assign IFR[3] = 0;
assign IFR[4] = 0;
assign IFR[5] = 0;
assign IFR[6] = 0;
assign IFR[7] = 0;		//save for later

always @(IFR)
begin
	casex ({IFR[IFR_RXD],IFR[IFR_TXDMT],IFR[IFR_TXDDONE]})
		3'b1xx: {I,Ivr} = 9'b000000000;
		3'b01x: {I,Ivr} = 9'b000000001;
		3'b001: {I,Ivr} = 9'b000000010;
		3'b000: {I,Ivr} = 9'b100000000;
		default:{I,Ivr} = 9'b100000000;
	endcase
end

//-------------------------------------
// write strobe decodes
//-------------------------------------

always @ (wd or A) begin
	if(wd && (A == UART_BRATE_L))
		wdBL = 1;
	else
		wdBL = 0;
end

always @ (wd or A) begin
	if(wd && (A == UART_BRATE_H))
		wdBH = 1;
	else
		wdBH = 0;
end

always @ (wd or A) begin
	if(wd && (A == UART_IER))
		wdIER = 1;
	else
		wdIER = 0;
end

always @ (wd or A) begin
	if(wd && (A == UART_DATA))
		wdTXD = 1;
	else
		wdTXD = 0;
end

assign wdClrIFR = wd & (A == UART_IFR);

//-------------------------------------
// Read Strobes
//
// A read strobe actually occures
// before a read operation...it is used
// to set things up for the actual
// read...for example...reading a FIFO
// is a clocked event, so we do a read
// strobe on the FIFO, and then we
// can actually see what is at the
// output of the FIFO with a uC.
//-------------------------------------

assign rdRFifo = rd & (A == UART_DATA);

//-------------------------------------
// Read Data MUX
//-------------------------------------

always @ (A or RFifoO or RDCount or TDCount or BaudL or BaudH or IER or REmpty or RFull or TEmpty or TFull or frame or overrun or IFR or Ivr) begin
	case (A)
		UART_DATA: O = RFifoO;
		UART_RSTATUS: O = {frame,overrun,REmpty,RFull,RDCount};
		UART_TSTATUS: O = {2'b0,TEmpty,TFull,TDCount};
		UART_BRATE_L: O = BaudL;
		UART_BRATE_H: O = BaudH;
		UART_IER:	  O = IER;
		UART_IFR:	  O = IFR;
		UART_IVR:	  O = Ivr;
		default: O = 8'h7c;	//just some wierd value for debug
	endcase
end

//-------------------------------------
// Interrupt Enable Register
//-------------------------------------

always @ (posedge clk or posedge reset) begin
	if(reset)
		IER <= 0;	//disable all interrupts
	else if(wdIER)
		IER <= D;	//write data into IER
end

//-------------------------------------
// Baud Rate Registers
//-------------------------------------

always @ (posedge clk or posedge reset) begin
	if(reset)
		BaudL <= 0;
	else if (wdBL)
		BaudL <= D;	//write data into register
end

always @ (posedge clk or posedge reset) begin
	if(reset)
		BaudH <= 0;
	else if (wdBH)
		BaudH <= D;	//write data into register
end

//-------------------------------------
// receive uart
//-------------------------------------



uartrx RxD (
    .dout(RDout), 
    .clk(clk), 
    .bclk(bclk), 
    .reset(reset), 
    .rxd(rxd),			 
    .frame(frame), 
    .overrun(overrun), 
    .ready(wdRFifo)
);

//-------------------------------------
// transmit uart
//-------------------------------------

uarttx TxD (
    .clock(clk), 
    .reset(reset), 
    .din(TFifoO), 
    .load(wdTxUart), 
    .brate(tbrate), 
    .txd(txd), 
    .done(done)
);

TBitRate TBitR (
    .clk(clk), 
    .reset(reset), 
    .sync(wdTxUart), 
    .R({BaudH,BaudL}), 
    .BR(tbrate)
);

   always@(posedge clk or posedge reset)
      if (reset) 
         TState <= TUART_STATE_IDLE;
      else
         (* FULL_CASE, PARALLEL_CASE *) case (TState)
            TUART_STATE_IDLE : begin
               if (!TEmpty)
                  TState <= TUART_STATE_RDFIFO;
               else
                  TState <= TUART_STATE_IDLE;
            end
            TUART_STATE_RDFIFO : begin
               TState <= TUART_STATE_WDUART;
            end
            TUART_STATE_WDUART : begin
               TState <= TUART_STATE_WAIT;
            end
            TUART_STATE_WAIT : begin
               if (done)
                  TState <= TUART_STATE_IDLE;
               else
                  TState <= TUART_STATE_WAIT;
            end
         endcase

   assign rdTFifo = (TState == TUART_STATE_RDFIFO);	//TUART_STATE_RDFIFO	
   assign wdTxUart =(TState == TUART_STATE_WDUART);	//TUART_STATE_WDUART

//-------------------------------------
// baud rate generator
//-------------------------------------

baudrategen BitRate (
    .clk(clk), 
    .reset(reset), 
    .R({BaudH,BaudL}), 
    .BR(bclk)
);

//-------------------------------------
// receive fifo
//-------------------------------------

fifo16x8 RFifo (
    .clk(clk), 
    .reset(reset), 
    .I(RDout), 
    .W(wdRFifo), 
    .R(rdRFifo), 
    .O(RFifoO), 
    .F(RFull), 
    .E(REmpty), 
    .N(RDCount)
    );

//-----------------------------------------
// Transmit Fifo
//-----------------------------------------

fifo16x8 TFifo (
    .clk(clk), 
    .reset(reset), 
    .I(D), 
    .W(wdTXD), 
    .R(rdTFifo), 
    .O(TFifoO), 
    .F(TFull), 
    .E(TEmpty), 
    .N(TDCount)
    );

endmodule
