//---------------------------------------------------------------------------------
// Uart Transmit Module
//
// Created March 6, 2005
// Author:James Patchell
// Copyleft (c) 2005 by James Patchell
// This code may be used freely in anyway you see fit...
// This is open source.
//
// This is a module that implements the transmit portion of an 8 bit uart.
// This is a very simple minded implementation in that that is all it does.
// There is no provision for any other word size, or the addition of parity,
// since in general I have absolutely no use for any of that.
//
// parameter:
//	din............8 bit bus of data that is to be sent out
//	load...........This signal write the data to a holding register and
//				sets a Ready flag to let the transmitter know that
//				there is data ready to go out.
//	clock..........system clock
//	reset..........system reset, returns uart to an idle stat and clears ready
//	brate..........signal to let the system know when it is time to shift a bit
//				(this is the baud rate).
//	txd............transmit data output
//	ready..........status: 0 indicates no data in holding register
//				status: 1 indicates holding register is full
// done...........goes true for one clock cycle when transmit is done
//------------------------------------------------------------------------------
// Revision History:
// Rev 1.01
//	April 17, 2005
//	Changed the shift register to shift the LSB out first
//----------------------------------------------------------------------------------

module uarttx(clock,reset,din,load,brate,txd,done);
	input clock;			// system clock
	input reset; 			// resets the registers
	input [7:0] din;		//data to be transmitted
	input load;				//loads the transmit register
	input brate;			// this is when the shift register is supposed to shift
	output txd;				// output data
	output done;			//indicates when transmit is done
	
	wire txd;				//transmit bit
	reg done;			//indicates when transmit is done
	
	reg [8:0] INT;		//nine bit shift register
	reg doshift;		//tells the shift register to shift its data
			

//-----------------------------------------------------------------
// Uart State machine
//
// it should be noted, the LSB is the first bit to
// be transmitted...the states would seem to imply
// that we start at BIT7...well, that is wrong
//-----------------------------------------------------------------

parameter [3:0] //synopsys enum STATE_TYPE
    UART_IDLE = 4'b0000,
    UART_STARTBIT = 4'b0001,
    UART_BIT7 = 4'b0010,
    UART_BIT6 = 4'b0011,
    UART_BIT5 = 4'b0100,
    UART_BIT4 = 4'b0101,
    UART_BIT3 = 4'b0110,
    UART_BIT2 = 4'b0111,
    UART_BIT1 = 4'b1000,
    UART_BIT0 = 4'b1001,
    UART_STOPBIT = 4'b1010;
 
// Declare current state and next state variables
reg [3:0] /* synopsys enum STATE_TYPE */ CS;
reg [3:0] /* synopsys enum STATE_TYPE */ NS;
 
//synopsys state_vector CS
 
    always @ (posedge clock or posedge reset)
    begin
       if (reset) CS <= UART_IDLE;
       else  CS <= NS;
    end
 
	always @ (CS or load or brate)
	begin 
		case (CS) //synopsys full_case
			UART_IDLE: begin
				if (load) begin
					NS <= UART_STARTBIT;
					doshift = 0;
				end
				else begin
					NS <= UART_IDLE;
					doshift = 0;
				end
			end 
			UART_STARTBIT: begin
				if(brate) begin
					NS <= UART_BIT7;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_STARTBIT;	//hold this state
					doshift = 0;
				end
			end
			UART_BIT7: begin
				if(brate) begin
					NS <= UART_BIT6;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT7;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT6: begin
				if(brate) begin
					NS <= UART_BIT5;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT6;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT5: begin
				if(brate) begin
					NS <= UART_BIT4;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT5;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT4: begin
				if(brate) begin
					NS <= UART_BIT3;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT4;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT3: begin
				if(brate)	begin
					NS <= UART_BIT2;		//go to next state
					doshift = 1;			//shift data out register
				end
				else	begin
					NS <= UART_BIT3;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT2: begin
				if(brate) begin
					NS <= UART_BIT1;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT2;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT1: begin
				if(brate) begin
					NS <= UART_BIT0;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT1;		//hold this state
					doshift = 0;
				end
			end
			UART_BIT0: begin
				if(brate) begin
					NS <= UART_STOPBIT;		//go to next state
					doshift = 1;			//shift data out register
				end
				else begin
					NS <= UART_BIT0;		//hold this state
					doshift = 0;
				end
			end
			UART_STOPBIT: begin
				if(brate) begin
					NS <= UART_IDLE;		//nothing more to do, so idle
					doshift = 0;			
				end
				else begin
					NS <= UART_STOPBIT;		//hold this state
					doshift = 0;			
				end
			end
			default: begin
				doshift = 0;			
				NS <= UART_IDLE;
			end 
		endcase
    end 

	always @ (brate or CS)
		if((CS == UART_STOPBIT) && brate)
			done = 1;
		else
			done = 0;
	
//---------------------------------------------------------------
// shift register
//
// shift register can do a load, and do a shift
//---------------------------------------------------------------

always @(posedge clock or posedge reset)
begin
	if(reset)
		INT <= 9'b111111111;	//reset transmit register to all 1's
	else if(load)
		INT <= {din,1'b0};	//load data and set start bit to 0
	else if (doshift)
		INT <= {1'b1,INT[8:1]};	//shift data, shift in 1's
end

assign txd = INT[0];

endmodule
